import json
from functools import partial

import anyio
import typer

from . import get_abuse_contacts
from .errors import InvalidAddressError, TimeoutError

app = typer.Typer()


@app.command()
def whois(
    address: str = typer.Argument(..., help="URL, domain, IP address or email address")
):
    try:
        contacts = anyio.run(partial(get_abuse_contacts, address))
        print(contacts.json(by_alias=True))  # noqa: T001
    except (InvalidAddressError, TimeoutError) as e:
        print(json.dumps({"error": str(e)}))  # noqa: T001


if __name__ == "__main__":
    app()



from functools import partial
from typing import Optional

import aiometer

from abuse_whois.matchers.shared_hosting import get_shared_hosting_provider
from abuse_whois.matchers.whois import get_contact_from_whois

from .errors import InvalidAddressError, TimeoutError
from .ip import resolve_ip_address
from .schemas import Contact, Contacts, WhoisRecord
from .utils import (
    get_hostname,
    get_registered_domain,
    is_domain,
    is_ip_address,
    is_supported_address,
)
from .whois import get_whois_record as _get_whois_record


async def get_whois_record(hostname: str) -> Optional[WhoisRecord]:
    try:
        return await _get_whois_record(hostname)
    except TimeoutError:
        return None


async def get_contact(domain_or_ip_address: Optional[str] = None) -> Optional[Contact]:
    if domain_or_ip_address is None:
        return None

    return await get_contact_from_whois(domain_or_ip_address)


async def get_registrar_and_hosting_provider_contacts(
    *, domain: Optional[str] = None, ip_address: Optional[str] = None
):
    values = [domain, ip_address]
    return await aiometer.run_all([partial(get_contact, value) for value in values])


async def get_abuse_contacts(address: str) -> Contacts:
    if not is_supported_address(address):
        raise InvalidAddressError(f"{address} is not supported type address")

    shared_hosting_provider: Optional[Contact] = None
    registrar: Optional[Contact] = None
    hosting_provider: Optional[Contact] = None

    hostname = get_hostname(address)

    domain: Optional[str] = None  # FQDN
    ip_address: Optional[str] = None
    registered_domain: Optional[str] = None
    whois_record: Optional[WhoisRecord] = None

    shared_hosting_provider = get_shared_hosting_provider(hostname)

    if is_domain(hostname):
        domain = hostname
        registered_domain = get_registered_domain(hostname)

        # get IP address by domain
        try:
            ip_address = await resolve_ip_address(hostname)
        except OSError:
            pass

    if is_ip_address(hostname):
        ip_address = hostname

    registrar, hosting_provider = await get_registrar_and_hosting_provider_contacts(
        domain=domain, ip_address=ip_address
    )

    # it will get cached result (if there is no TimeoutError)
    whois_record = await get_whois_record(hostname)

    return Contacts(
        address=address,
        hostname=hostname,
        ip_address=ip_address,
        registered_domain=registered_domain,
        shared_hosting_provider=shared_hosting_provider,
        registrar=registrar,
        hosting_provider=hosting_provider,
        whois_record=whois_record,
    )



from .main import get_abuse_contacts  # noqa: F401

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata


__version__ = importlib_metadata.version(__name__)



from humps import camelize
from pydantic import BaseModel


class APIModel(BaseModel):
    class Config:
        orm_mode = True
        alias_generator = camelize
        allow_population_by_field_name = True



from fastapi import APIRouter, HTTPException

from abuse_whois import get_abuse_contacts, schemas
from abuse_whois.errors import InvalidAddressError

router = APIRouter(prefix="/whois")


@router.post("/", response_model=schemas.Contacts)
async def whois(query: schemas.Query):
    try:
        return await get_abuse_contacts(query.address)
    except InvalidAddressError as e:
        raise HTTPException(400, detail=str(e))
    except TimeoutError as e:
        raise HTTPException(500, detail=str(e))



from typing import List

from pydantic import Field

from abuse_whois.schemas import BaseRule, Contact
from abuse_whois.utils import is_included_in_base_domains
from abuse_whois.whois import get_whois_record


class WhoisRule(BaseRule):
    contact: Contact
    keywords: List[str] = Field(default_factory=list)
    base_domains: List[str] = Field(default_factory=list)

    async def match(self, hostname: str) -> bool:
        if is_included_in_base_domains(self.base_domains, hostname):
            return True

        try:
            whois_record = await get_whois_record(hostname)
        except Exception:
            return False

        for keyword in self.keywords:
            if keyword in whois_record.raw_text:
                return True

        return False



import pathlib
from functools import lru_cache
from typing import Dict, List, Optional, Union, cast
from urllib.parse import urlparse

import tldextract
import yaml
from email_validator import EmailNotValidError, validate_email
from pydantic import BaseModel, ValidationError
from pydantic.networks import AnyHttpUrl

from abuse_whois import settings


class URLModel(BaseModel):
    url: AnyHttpUrl


def is_ip_address(v: str) -> bool:
    try:
        model = URLModel(url=f"http://{v}")
        return model.url.host_type in ["ipv4", "ipv6"]
    except ValidationError:
        return False


def is_domain(v: str) -> bool:
    if len(v.split(".")) == 1:
        return False

    if "@" in v:
        return False

    try:
        model = URLModel(url=f"http://{v}")
        return model.url.host_type in ["domain", "int_domain"]
    except ValidationError:
        return False


def is_url(v: str) -> bool:
    if not v.startswith(("http://", "https://")):
        return False

    try:
        URLModel(url=v)
        return True
    except ValidationError:
        return False


def is_email(v: str) -> bool:
    try:
        validate_email(v, check_deliverability=False)
        return True
    except EmailNotValidError:
        return False


def is_supported_address(v: str) -> bool:
    if is_domain(v) or is_ip_address(v) or is_email(v) or is_url(v):
        return True

    return False


@lru_cache(maxsize=settings.WHOIS_LOOKUP_CACHE_SIZE)
def get_registered_domain(v: str) -> Optional[str]:
    parsed = tldextract.extract(v)

    if parsed.registered_domain == "":
        return None

    return parsed.registered_domain


def get_hostname(value: str) -> str:
    if is_ip_address(value) or is_domain(value):
        return value

    if is_email(value):
        value = f"http://{value}"

    parsed = urlparse(value)
    return parsed.hostname or value


def load_yaml(path: Union[str, pathlib.Path]) -> Dict:
    with open(path) as f:
        return cast(Dict, yaml.safe_load(f))


def is_included_in_base_domains(base_domains: List[str], hostname: str) -> bool:
    if hostname in base_domains:
        return True

    for base_domain in base_domains:
        if hostname.endswith(f".{base_domain}"):
            return True

    return False



import re
from typing import Optional, Pattern

from abuse_whois.schemas import Contact, WhoisRecord
from abuse_whois.utils import is_email
from abuse_whois.whois import get_whois_record

from .rules import load_rules


def get_whois_abuse_contact_by_regexp(
    record: WhoisRecord, *, abuse_email_pattern: Pattern = r"abuse@[a-z0-9\-.]+"
) -> Optional[Contact]:
    provider = record.registrar or ""

    matches = re.findall(abuse_email_pattern, record.raw_text)
    if len(matches) == 0:
        return None

    # returns the email address in the bottom
    matches.reverse()
    for match in matches:
        if is_email(str(match)):
            return Contact(provider=provider, address=str(match))

    return None


def get_whois_abuse_contact(record: WhoisRecord) -> Optional[Contact]:
    provider = record.registrar
    email: Optional[str] = None

    # check email format for just in case
    if is_email(record.abuse.email or ""):
        email = record.abuse.email

    if email is None:
        # fallback to regexp based search
        contact = get_whois_abuse_contact_by_regexp(record)
        if contact is None:
            return None

        provider = contact.provider
        email = contact.address

    # use email's domain as a provider
    if is_email(provider or ""):
        provider = email.split("@")[-1]

    if provider is None or email is None:
        return None

    return Contact(provider=provider, address=email)


async def get_contact_from_whois(
    hostname: str,
) -> Optional[Contact]:
    rules = load_rules()
    for rule in rules:
        if await rule.match(hostname):
            return rule.contact

    # Use whois registrar & abuse data
    try:
        whois_record = await get_whois_record(hostname)
    except Exception:
        return None

    return get_whois_abuse_contact(whois_record)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['abuse_whois',
 'abuse_whois.api',
 'abuse_whois.api.endpoints',
 'abuse_whois.matchers',
 'abuse_whois.matchers.shared_hosting',
 'abuse_whois.matchers.whois',
 'abuse_whois.schemas']

package_data = \
{'': ['*'],
 'abuse_whois.matchers.shared_hosting': ['rules/*'],
 'abuse_whois.matchers.whois': ['rules/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aiometer>=0.3.0,<0.4.0',
 'asyncache>=0.1.1,<0.2.0',
 'asyncer>=0.0.1,<0.0.2',
 'cachetools>=5.2.0,<6.0.0',
 'email-validator>=1.2.1,<2.0.0',
 'fastapi>=0.78.0,<0.79.0',
 'loguru>=0.6.0,<0.7.0',
 'pydantic>=1.9.1,<2.0.0',
 'pyhumps>=3.7.2,<4.0.0',
 'tldextract>=3.3.0,<4.0.0',
 'typer>=0.4.1,<0.5.0',
 'whois-parser>=0.1.4,<0.2.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=4.11.4,<5.0.0',
                             'typing-extensions>=3.10,<4.0'],
 'api': ['gunicorn>=20.1.0,<21.0.0', 'uvicorn[standard]>=0.18.1,<0.19.0']}

entry_points = \
{'console_scripts': ['abuse_whois = abuse_whois.cli:app']}

setup_kwargs = {
    'name': 'abuse-whois',
    'version': '0.4.0',
    'description': 'Find where to report a domain for abuse',
    'long_description': '# abuse_whois\n\n[![PyPI version](https://badge.fury.io/py/abuse-whois.svg)](https://badge.fury.io/py/abuse-whois)\n[![Python CI](https://github.com/ninoseki/abuse_whois/actions/workflows/test.yml/badge.svg)](https://github.com/ninoseki/abuse_whois/actions/workflows/test.yml)\n[![Coverage Status](https://coveralls.io/repos/github/ninoseki/abuse_whois/badge.svg?branch=main)](https://coveralls.io/github/ninoseki/abuse_whois?branch=main)\n\nYet another way to find where to report an abuse.\n\n![img](./images/overview.jpg)\n\nThis tool is highly inspired from the following libraries:\n\n- https://github.com/bradleyjkemp/abwhose\n- https://github.com/certsocietegenerale/abuse_finder\n\n## Requirements\n\n- Python 3.7+\n- whois\n\n## Installation\n\n```bash\npip install abuse_whois\n\n# or if you want to use built-in REST API\npip install abuse_whois[api]\n```\n\n## Usage\n\n### As a library\n\n```python\nfrom abuse_whois import get_abuse_contacts\n\nget_abuse_contacts("1.1.1.1")\nget_abuse_contacts("github.com")\nget_abuse_contacts("https://github.com")\nget_abuse_contacts("foo@example.com")\n```\n\n### As a CLI tool\n\n```bash\n$ abuse_whois 1.1.1.1 | jq .\n```\n\n### As a REST API\n\n```bash\n$ uvicorn abuse_whois.api.app:app\nINFO:     Started server process [2283]\nINFO:     Waiting for application startup.\nINFO:     Application startup complete.\nINFO:     Uvicorn running on http://127.0.0.1:8000 (Press CTRL+C to quit)\n\n$ http localhost:8000/api/whois/ address=https://github.com\n```\n\n### With Docker\n\n```bash\ngit clone https://github.com/ninoseki/abuse_whois\ncd abuse_whois\ndocker build . -t abuse-whois\ndocker run -i -d -p 8000:8000 abuse-whois\n```\n\n## Settings\n\nAll settings can be done via environment variables or `.env` file.\n\n| Name                         | Type | Default | Desc.                                           |\n|------------------------------|------|---------|-------------------------------------------------|\n| WHOIS_LOOKUP_TIMEOUT         | int  | 10      | Timeout value for whois lookup (seconds)        |\n| WHOIS_LOOKUP_CACHE_SIZE      | int  | 1024    | Cache size for whois lookup                     |\n| WHOIS_LOOKUP_CACHE_TTL       | int  | 3600    | Cache TTL value for whois lookup (seconds)      |\n| IP_ADDRESS_LOOKUP_TIMEOUT    | int  | 10      | Timeout value for IP address lookup (seconds)   |\n| IP_ADDRESS_LOOKUP_CACHE_SIZE | int  | 1024    | Cache size for IP address lookup                |\n| IP_ADDRESS_LOOKUP_CACHE_TTL  | int  | 3600    | Cache TTL value for IP address lookup (seconds) |\n\n## Contributions\n\n`abuse_whois` works based on a combination of static rules and a parsing result of whois response.\n\n- Rules:\n  - [Registrar and hosting provider](https://github.com/ninoseki/abuse_whois/wiki/Registrar-and-Hosting-Provider)\n  - [Shared hosting provider](https://github.com/ninoseki/abuse_whois/wiki/Shared-Hosting)\n\nPlease submit a PR (or submit a feature request) if you find something missing.\n',
    'author': 'Manabu Niseki',
    'author_email': 'manabu.niseki@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ninoseki/abuse_whois',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import pathlib
from functools import lru_cache
from typing import List

from abuse_whois.utils import load_yaml

from .rule import SharedHostingRule


@lru_cache(maxsize=1)
def load_rules() -> List[SharedHostingRule]:
    rules: List[SharedHostingRule] = []
    for path in pathlib.Path(__file__).parent.glob("./rules/*.yaml"):
        data = load_yaml(path)
        rules.append(SharedHostingRule.parse_obj(data))

    return rules



from typing import List

from pydantic import Field

from abuse_whois.schemas import BaseRule, Contact
from abuse_whois.utils import is_included_in_base_domains


class SharedHostingRule(BaseRule):
    contact: Contact
    base_domains: List[str] = Field(default_factory=list)

    def match(self, hostname: str) -> bool:
        return is_included_in_base_domains(self.base_domains, hostname)



import asyncio
import functools
import shlex
import warnings

from asyncache import cached
from cachetools import TTLCache
from whois_parser import WhoisParser

from . import schemas, settings
from .errors import TimeoutError
from .utils import get_registered_domain, is_ip_address

# Ignore dateparser warnings regarding pytz
# ref. https://github.com/scrapinghub/dateparser/issues/1013
warnings.filterwarnings(
    "ignore",
    message="The localize method is no longer necessary, as this time zone supports the fold attribute",
)


@functools.lru_cache(maxsize=1)
def get_whois_parser() -> WhoisParser:
    return WhoisParser()


def parse(raw_text: str, hostname: str) -> schemas.WhoisRecord:
    parser = get_whois_parser()
    record = parser.parse(raw_text, hostname=hostname)

    return schemas.WhoisRecord.parse_obj(record.to_dict())


@cached(
    cache=TTLCache(
        maxsize=settings.WHOIS_LOOKUP_CACHE_SIZE, ttl=settings.WHOIS_LOOKUP_CACHE_TTL
    )
)
async def get_whois_record(
    hostname: str, *, timeout: int = settings.WHOIS_LOOKUP_TIMEOUT
) -> schemas.WhoisRecord:
    if not is_ip_address(hostname):
        hostname = get_registered_domain(hostname) or hostname

    # open a new process for "whois" command
    cmd = f"whois {shlex.quote(hostname)}"
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    try:
        # block for query_result
        query_result_, _ = await asyncio.wait_for(proc.communicate(), timeout=timeout)
        query_result = query_result_.decode(errors="ignore")
        query_result = query_result.strip()
    except asyncio.TimeoutError:
        raise TimeoutError(f"{timeout} seconds have passed but there is no response")

    return parse(query_result, hostname)



import socket
from contextlib import contextmanager

from asyncer import asyncify
from cachetools import TTLCache, cached

from . import settings
from .errors import TimeoutError


@contextmanager
def socket_with_timeout(timeout: float):
    old_timeout = socket.getdefaulttimeout()

    try:
        socket.setdefaulttimeout(timeout)
        yield socket
    except (socket.timeout, ValueError):
        raise TimeoutError(f"{timeout} seconds have passed but there is no response")
    finally:
        socket.setdefaulttimeout(old_timeout)


@cached(
    cache=TTLCache(
        maxsize=settings.IP_ADDRESS_LOOKUP_CACHE_SIZE,
        ttl=settings.IP_ADDRESS_LOOKUP_CACHE_TTL,
    )
)
def _resolve_ip_address(
    hostname: str, *, timeout: int = settings.IP_ADDRESS_LOOKUP_TIMEOUT
) -> str:
    with socket_with_timeout(float(timeout)):
        ip = socket.gethostbyname(hostname)
        return ip


resolve_ip_address = asyncify(_resolve_ip_address)



from datetime import datetime
from typing import List, Optional, Union

from pydantic import Field

from .api_model import APIModel


class Contact(APIModel):
    organization: Optional[str] = None
    email: Optional[str] = None
    name: Optional[str] = None
    telephone: Optional[str] = None


class Abuse(APIModel):
    email: Optional[str] = None
    telephone: Optional[str] = None


class WhoisRecord(APIModel):
    raw_text: str

    tech: Contact
    admin: Contact
    registrant: Contact
    abuse: Abuse

    statuses: List[str] = Field(default_factory=list)
    name_servers: List[str] = Field(default_factory=list)

    domain: Optional[str] = None
    registrar: Optional[str] = None

    expires_at: Optional[Union[datetime, str]] = None
    registered_at: Optional[Union[datetime, str]] = None
    updated_at: Optional[Union[datetime, str]] = None
